/*
 * The contents of this file are subject to the Mozilla Public License 
 * Version 1.1 (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific language governing rights and limitations 
 * under the License.
 * 
 * The Original Code is the NRL Parser, released 28 April 2006, 
 * Copyright (c) Christian Nentwich. The Initial Developer of the 
 * Original Code is Christian Nentwich. Portions created by contributors 
 * identified in the NOTICES file are Copyright (c) the contributors. 
 * All Rights Reserved. 
 */
package net.sourceforge.nrl.parser.model.uml2;

import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import net.sourceforge.nrl.parser.model.AbstractAttribute;
import net.sourceforge.nrl.parser.model.IDataType;
import net.sourceforge.nrl.parser.model.IModelElement;
import net.sourceforge.nrl.parser.model.PrimitiveTypeFactory;

import org.eclipse.uml2.uml.Classifier;
import org.eclipse.uml2.uml.DataType;
import org.eclipse.uml2.uml.EnumerationLiteral;
import org.eclipse.uml2.uml.Generalization;
import org.eclipse.uml2.uml.Model;
import org.eclipse.uml2.uml.MultiplicityElement;
import org.eclipse.uml2.uml.NamedElement;
import org.eclipse.uml2.uml.PrimitiveType;
import org.eclipse.uml2.uml.Property;
import org.eclipse.uml2.uml.Type;
import org.eclipse.uml2.uml.TypedElement;

/**
 * Extension of a standard attribute that holds a UML2 typed element reference.
 * 
 * @author Christian Nentwich
 */
public class UML2Attribute extends AbstractAttribute {

	// The UML named element we're wrapping
	private NamedElement umlNamedElement;

	/**
	 * Initialise the attribute using a UML2 named element object. The type of
	 * the attribute will remain unresolved until {@link #resolve(Map)} is
	 * called.
	 * 
	 * @param namedElement the element
	 */
	public UML2Attribute(NamedElement namedElement) {
		super(namedElement.getName());
		
		setName(cleanUpName(getName()));
		this.umlNamedElement = namedElement;
	}

	/**
	 * Given the attribute name, clean it up by removing illegal characters like
	 * dots.
	 * 
	 * @param name the name
	 * @return the cleaned up name
	 */
	protected String cleanUpName(String name) {
		StringBuffer result = new StringBuffer();
		
		StringTokenizer tokenizer = new StringTokenizer(name, ".");
		int count = 0;
		while (tokenizer.hasMoreTokens()) {
			String token = tokenizer.nextToken();
			
			if (count == 0) {
				result.append(token);
			} else {
				result.append(Character.toUpperCase(token.charAt(0)));
				if (token.length() > 1)
					result.append(token.substring(1));
			}
			
			count++;
		}
		
		return result.toString();
	}

	/**
	 * Extract the documentation from the UML attribute.
	 */
	public List<String> getDocumentation() {
		return UML2Helper.extractComments(umlNamedElement);
	}

	public int getMinOccurs() {
		if (umlNamedElement instanceof MultiplicityElement) {
			int lower = ((MultiplicityElement) umlNamedElement).getLower();
			if (lower == -1)
				return UNBOUNDED;
			else
				return lower;
		}
		return 1;
	}

	public int getMaxOccurs() {
		if (umlNamedElement instanceof MultiplicityElement) {
			int upper = ((MultiplicityElement) umlNamedElement).getUpper();
			if (upper == -1)
				return UNBOUNDED;
			else
				return upper;
		}
		return 1;
	}

	public boolean isStatic() {
		if (umlNamedElement instanceof EnumerationLiteral)
			return true;
		if ((umlNamedElement instanceof Property))
			return ((Property) umlNamedElement).isStatic();
		return false;
	}

	/**
	 * Return the UML2 Property this attribute was created from.
	 * 
	 * @return the property
	 */
	public NamedElement getUML2() {
		return umlNamedElement;
	}
	
	@Override
	public Object getUserData(String key) {
		if (IUML2UserData.UML2_ELEMENT.equals(key))
			return getUML2();
		return super.getUserData(key);
	}
	

	/**
	 * Resolve the types of the attribute
	 */
	protected void resolve(Map<Classifier, IModelElement> umlClassToModelElement) {
		if ((umlNamedElement instanceof TypedElement)
				&& ((TypedElement) umlNamedElement).getType() != null) {

			TypedElement umlTypedElement = (TypedElement) umlNamedElement;

			Type type = umlTypedElement.getType();

			// FIX: for anonymous data types, use the supertype
			if (type != null && type.getName() != null
					&& type.getName().startsWith("anonymous") && type instanceof DataType) {

				if (((DataType) type).getGeneralizations().size() > 0) {
					type = ((Generalization) ((DataType) type).getGeneralizations()
							.get(0)).getGeneral();
				}
			}

			// Check if the type is a basic UML2 data type
			if (type != null && type instanceof PrimitiveType
					&& type.eContainer() instanceof Model
					&& ((Model) type.eContainer()).getName().toLowerCase().equals("uml2")) {

				String name = umlTypedElement.getType().getName().toLowerCase();
				IDataType typeToSet = PrimitiveTypeFactory.getInstance().getType(name);
				if (typeToSet != null)
					setType(typeToSet);
			}

			IModelElement resultType = (IModelElement) umlClassToModelElement.get(type);
			if (resultType != null)
				setType(resultType);
		}
	}

}
