/*
 * The contents of this file are subject to the Mozilla Public License Version 1.1 (the "License");
 * you may not use this file except in compliance with the License. You may obtain a copy of the
 * License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the License for the specific language governing rights
 * and limitations under the License.
 * 
 * The Original Code is the NRL Parser, released 28 April 2006, Copyright (c) Christian Nentwich.
 * The Initial Developer of the Original Code is Christian Nentwich. Portions created by
 * contributors identified in the NOTICES file are Copyright (c) the contributors. All Rights
 * Reserved.
 */
package net.sourceforge.nrl.parser;

import java.io.FileInputStream;
import java.io.InputStreamReader;

import net.sourceforge.nrl.parser.ast.IModelReference;
import net.sourceforge.nrl.parser.ast.IRuleFile;
import net.sourceforge.nrl.parser.ast.constraints.IBinaryPredicate;
import net.sourceforge.nrl.parser.ast.constraints.IConstraintRuleDeclaration;
import net.sourceforge.nrl.parser.ast.impl.SyntaxErrorException;
import net.sourceforge.nrl.parser.model.IPackage;
import net.sourceforge.nrl.parser.model.ModelCollection;

/**
 * Test the main parser class.
 * 
 * @author Christian Nentwich
 */
public class NRLParserTest extends NRLParserTestCase {

	/**
	 * Test parsing and model resolution.
	 */
	public void testParse() throws Exception {
		INRLParser parser = new NRLParser();

		IRuleFile ruleFile = parser
				.parse(new FileInputStream("testdata/parsing/model-correct.nrl"));
		assertEquals(0, parser.getErrors().size());
		assertNotNull(ruleFile);

		IPackage model = getBasicModel();
		assertNotNull(model);

		ModelCollection models = new ModelCollection();
		models.addModelPackage(model);

		parser.resolveModelReferences(ruleFile, models);
		assertEquals(0, parser.getErrors().size());

		// Check for correct resolution
		assertEquals("Trade", ruleFile.getRuleById("expr-1").getContext().getName());
	}

	/*
	 * Check if illegal chars are reported using syntax errors
	 */
	public void testParse_illegalChar() throws Exception {
		INRLParser parser = new NRLParser();
		try {
			parser.parse(new FileInputStream("testdata/parsing/illegal-char.nrl"));
			fail("should throw exception");
		} catch (SyntaxErrorException e) {
			assertEquals(9, e.getSyntaxError().getLine());
			assertEquals(13, e.getSyntaxError().getColumn());
		}
	}

	/**
	 * Test parsing UTF8 files
	 */
	public void testParse_UTF8() throws Exception {
		INRLParser parser = new NRLParser();
		IRuleFile ruleFile = null;
		try {
			ruleFile = parser.parse(new InputStreamReader(new FileInputStream("testdata/parsing/utf8.nrl"), "utf-8"));			
		} catch (SyntaxErrorException e) {
			System.err.println(e.getSyntaxError().getLine()+" " + e);
			fail(e.getMessage());
		}
		assertEquals(0, parser.getErrors().size());
		assertNotNull(ruleFile);
		assertNotNull(ruleFile.getRuleById("çãüäö"));
		assertEquals("çãüäö",
				((IModelReference) ((IBinaryPredicate) ((IConstraintRuleDeclaration) ruleFile
						.getRuleById("çãüäö")).getConstraint()).getLeft()).getOriginalString());

		assertNotNull(ruleFile.getRuleById("日本語"));
		assertEquals("日本語",
				((IModelReference) ((IBinaryPredicate) ((IConstraintRuleDeclaration) ruleFile
						.getRuleById("日本語")).getConstraint()).getLeft()).getOriginalString());
	}

	/**
	 * Test parsing UTF16 files
	 */
	public void testParse_UTF16() throws Exception {
		INRLParser parser = new NRLParser();
		IRuleFile ruleFile = parser.parse(new InputStreamReader(new FileInputStream(
				"testdata/parsing/utf16.nrl"), "utf-16"));
		assertEquals(0, parser.getErrors().size());
		assertNotNull(ruleFile);
		assertNotNull(ruleFile.getRuleById("çãüäö"));
		assertEquals("çãüäö",
				((IModelReference) ((IBinaryPredicate) ((IConstraintRuleDeclaration) ruleFile
						.getRuleById("çãüäö")).getConstraint()).getLeft()).getOriginalString());

		assertNotNull(ruleFile.getRuleById("日本語"));
		assertEquals("日本語",
				((IModelReference) ((IBinaryPredicate) ((IConstraintRuleDeclaration) ruleFile
						.getRuleById("日本語")).getConstraint()).getLeft()).getOriginalString());
	}
}
