package net.sourceforge.nrl.parser.ast.impl;

import java.io.File;

import javax.xml.bind.JAXBContext;

import net.sourceforge.nrl.parser.IStatusCode;
import net.sourceforge.nrl.parser.NRLParserTestCase;
import net.sourceforge.nrl.parser.SemanticError;
import net.sourceforge.nrl.parser.ast.IRuleFile;
import net.sourceforge.nrl.parser.ast.constraints.IConstraintRuleDeclaration;
import net.sourceforge.nrl.parser.ast.constraints.IOperatorInvocation;
import net.sourceforge.nrl.parser.ast.impl.AntlrOperatorResolverVisitor;
import net.sourceforge.nrl.parser.model.ModelCollection;
import net.sourceforge.nrl.parser.operators.IOperators;
import net.sourceforge.nrl.parser.operators.Operators;
import net.sourceforge.nrl.parser.operators.XmlOperatorPersistence;
import net.sourceforge.nrl.parser.operators.XmlOperatorPersistence.LoadedVersion;

public class AntlrOperatorResolverVisitorTest extends NRLParserTestCase {

	/**
	 * Test valid resolution.
	 * 
	 * @throws Exception
	 */
	public void testResolve_ValidNoParams() throws Exception {
		IRuleFile ruleFile = createTestFile("[Operator two]");

		ModelCollection models = new ModelCollection();
		models.addModelPackage(getBasicModel());
		
		IOperators operators = new Operators();
		assertEquals(0, operators.resolveModelReferences(models).size());
		
		XmlOperatorPersistence loader = new XmlOperatorPersistence(JAXBContext
				.newInstance("net.sourceforge.nrl.parser.jaxb14"), JAXBContext
				.newInstance("net.sourceforge.nrl.parser.jaxb15"));
		LoadedVersion version = loader.load(new File("testdata/operators/operators.xml"));
		assertEquals(LoadedVersion.Version14, version);

		operators = loader.getOperators();
		assertNotNull(operators);

		AntlrOperatorResolverVisitor visitor = new AntlrOperatorResolverVisitor(
				new IOperators[] { operators });
		ruleFile.accept(visitor);
		assertEquals(0, visitor.getErrors().size());

		IOperatorInvocation invocation = (IOperatorInvocation) ((IConstraintRuleDeclaration) ruleFile
				.getRuleById("test-rule")).getConstraint();
		assertNotNull(invocation.getOperator());
		assertTrue(invocation.getOperator() == operators.getOperator("Operator two"));
	}

	/**
	 * Test valid resolution with no parameters.
	 * 
	 * @throws Exception
	 */
	public void testResolve_ValidWithParams() throws Exception {
		IRuleFile ruleFile = createTestFile("'ab' [Operator one] 'cd'");

		XmlOperatorPersistence loader = new XmlOperatorPersistence(JAXBContext
				.newInstance("net.sourceforge.nrl.parser.jaxb14"), JAXBContext
				.newInstance("net.sourceforge.nrl.parser.jaxb15"));
		LoadedVersion version = loader.load(new File("testdata/operators/operators.xml"));
		assertEquals(LoadedVersion.Version14, version);

		IOperators operators = loader.getOperators();
		assertNotNull(operators);
		
		AntlrOperatorResolverVisitor visitor = new AntlrOperatorResolverVisitor(
				new IOperators[] { operators });
		ruleFile.accept(visitor);
		assertEquals(0, visitor.getErrors().size());

		IOperatorInvocation invocation = (IOperatorInvocation) ((IConstraintRuleDeclaration) ruleFile
				.getRuleById("test-rule")).getConstraint();
		assertNotNull(invocation.getOperator());
		assertTrue(invocation.getOperator() == operators.getOperator("Operator one"));
	}

	/**
	 * Test error detection of invalid names
	 * 
	 * @throws Exception
	 */
	public void testResolve_InvalidName() throws Exception {
		IRuleFile ruleFile = createTestFile("'ab' [Operator one] 'cd'");

		// Empty collection
		Operators operators = new Operators();

		AntlrOperatorResolverVisitor visitor = new AntlrOperatorResolverVisitor(
				new IOperators[] { operators });
		ruleFile.accept(visitor);
		assertEquals(1, visitor.getErrors().size());

		assertEquals(IStatusCode.OPERATOR_UNKNOWN, ((SemanticError) visitor.getErrors()
				.get(0)).getStatusCode());
	}

	/**
	 * Parameter mismatch. Operator requires two parameters, we pass one.
	 * 
	 * @throws Exception
	 */
	public void testResolve_Parameters() throws Exception {
		IRuleFile ruleFile = createTestFile("[Operator one] 'cd'");

		XmlOperatorPersistence loader = new XmlOperatorPersistence(JAXBContext
				.newInstance("net.sourceforge.nrl.parser.jaxb14"), JAXBContext
				.newInstance("net.sourceforge.nrl.parser.jaxb15"));
		LoadedVersion version = loader.load(new File("testdata/operators/operators.xml"));
		assertEquals(LoadedVersion.Version14, version);

		IOperators operators = loader.getOperators();
		assertNotNull(operators);

		AntlrOperatorResolverVisitor visitor = new AntlrOperatorResolverVisitor(
				new IOperators[] { operators });
		ruleFile.accept(visitor);
		assertEquals(1, visitor.getErrors().size());

		assertEquals(IStatusCode.OPERATOR_PARAMETER_MISMATCH, ((SemanticError) visitor
				.getErrors().get(0)).getStatusCode());
	}

}
