/*
 * The contents of this file are subject to the Mozilla Public License 
 * Version 1.1 (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific language governing rights and limitations 
 * under the License.
 * 
 * The Original Code is the NRL Parser, released 28 April 2006, 
 * Copyright (c) Christian Nentwich. The Initial Developer of the 
 * Original Code is Christian Nentwich. Portions created by contributors 
 * identified in the NOTICES file are Copyright (c) the contributors. 
 * All Rights Reserved. 
 */
package net.sourceforge.nrl.parser.model.uml2;

import java.io.File;
import java.util.Iterator;

import net.sourceforge.nrl.parser.NRLParserTestCase;
import net.sourceforge.nrl.parser.model.IAttribute;
import net.sourceforge.nrl.parser.model.IClassifier;
import net.sourceforge.nrl.parser.model.IDataType;
import net.sourceforge.nrl.parser.model.IModelElement;
import net.sourceforge.nrl.parser.model.IPackage;

/**
 * Test the entire model loader. This tests all the model classes again, and also tests if
 * resolution works correctly.
 * 
 * @author Christian Nentwich
 */
public class UML2ModelLoaderTest extends NRLParserTestCase {

	/**
	 * Load a sample model (testdata/basicmodel.uml2) and check if all resolution worked.
	 */
	public void testLoad() throws Exception {
		UML2ModelLoader loader = new UML2ModelLoader();

		File file = new File("testdata/basicmodel.uml");
		IPackage model = loader.load(getResourceForFile(file), file);

		for (Iterator<String> iter = loader.getWarnings().iterator(); iter.hasNext();)
			System.out.println(iter.next());

		// Check the top package and model lookup functions
		assertEquals("BasicModel", model.getName());
		assertNotNull(model.getElementByName("Main", false));
		assertNull(model.getParent());
		assertNotNull(model.getElementByName("Trade", true));
		assertNull(model.getElementByName("Trade", false));
		assertNotNull(model.getElementByName("IRSwap", true));

		// Check the "Main" package
		IPackage main = (IPackage) model.getElementByName("Main", false);
		assertEquals(7, main.getSize());
		assertTrue(main.getElementByName("Trade", false) instanceof IClassifier);
		assertTrue(main.getElementByName("TradeHeader", false) instanceof IClassifier);
		assertTrue(main.getElementByName("Date", false) instanceof IDataType);
		assertNotNull(main.getUserData(IUML2UserData.UML2_ELEMENT));

		IClassifier trade = (IClassifier) model.getElementByName("Trade", true);
		IClassifier irSwap = (IClassifier) model.getElementByName("IRSwap", true);
		assertTrue(irSwap.getParent() == trade);
		assertTrue(trade.getDescendants(true).get(0) == irSwap);
		assertTrue(!trade.isSupplementary());
		assertNotNull(trade.getUserData(IUML2UserData.UML2_ELEMENT));

		assertEquals(IModelElement.ElementType.Classifier, trade.getElementType());
		assertEquals(IModelElement.ElementType.Classifier, irSwap.getElementType());

		// Check that trade has all attributes
		assertNotNull(trade.getAttributeByName("tradeDate", false));
		assertNotNull(trade.getAttributeByName("tradeheader", false));
		assertNotNull(trade.getAttributeByName("terminationDate", false));
		assertNotNull(trade.getAttributeByName("description", false));

		assertEquals(0, trade.getAttributeByName("description", false).getMinOccurs());
		assertEquals(IAttribute.UNBOUNDED, trade.getAttributeByName("description", false)
				.getMaxOccurs());

		assertNotNull(trade.getAttributeByName("description", false).getUserData(
				IUML2UserData.UML2_ELEMENT));

		// Check that the attribute types are resolved
		assertEquals("Date", trade.getAttributeByName("tradeDate", false).getType().getName());
		assertEquals("TradeHeader", trade.getAttributeByName("tradeheader", false).getType()
				.getName());
		assertEquals("String", trade.getAttributeByName("id", false).getType().getName());
		assertEquals("String", trade.getAttributeByName("description", false).getType().getName());

		// Check that IRSwap has inherited attributes
		assertNotNull(trade.getAttributeByName("tradeDate", false));
		assertNull(irSwap.getAttributeByName("tradeDate", false));
		assertNotNull(irSwap.getAttributeByName("tradeDate", true));
		assertNull(irSwap.getAttributeByName("tradeheader", false));
		assertNotNull(irSwap.getAttributeByName("tradeheader", true));

		// Check that the enumeration is there
		IClassifier _enum = (IClassifier) model.getElementByName("FixFloatEnum", true);
		assertNotNull(_enum);
		assertEquals(2, _enum.getAttributes(false).size());

		assertTrue(_enum == _enum.getAttributeByName("FIXED", true).getType());
		assertTrue(_enum == _enum.getAttributeByName("FLOAT", true).getType());
		assertEquals(1, _enum.getAttributeByName("FIXED", false).getMinOccurs());
		assertEquals(1, _enum.getAttributeByName("FIXED", false).getMaxOccurs());
		assertEquals(IModelElement.ElementType.Enumeration, _enum.getElementType());

		// Check that supplementary info is identified
		IModelElement str = model.getElementByName("String", true);
		assertNotNull(str);
		assertTrue(str.isSupplementary());

		// Test assignable from
		assertTrue(model.getElementByName("Trade", true).isAssignableFrom(
				model.getElementByName("Trade", true)));
		assertTrue(model.getElementByName("Trade", true).isAssignableFrom(
				model.getElementByName("IRSwap", true)));
		assertTrue(!model.getElementByName("IRSwap", true).isAssignableFrom(
				model.getElementByName("Trade", true)));

		// Test anonymous type
		IClassifier leg = (IClassifier) model.getElementByName("IRLeg", true);
		IAttribute attr = leg.getAttributeByName("ids", false);
		// assertNotNull(attr);
		// assertNotNull(attr.getType());
		//		
		// attr = ((IClassifier)attr.getType()).getAttributeByName("id", false);
		// assertNotNull(attr);
		// assertNotNull(attr.getType());

		// Test anonymous simple type
		leg = (IClassifier) model.getElementByName("IRLeg", true);
		attr = leg.getAttributeByName("name", false);
		assertNotNull(attr);
		assertNotNull(attr.getType());

		// Test attribute character processing (dotted.attribute in the model)
		attr = leg.getAttributeByName("dottedAttribute", false);
		assertNotNull(attr);
		assertNotNull(attr.getType());
	}
}
