/*
 * The contents of this file are subject to the Mozilla Public License 
 * Version 1.1 (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific language governing rights and limitations 
 * under the License.
 * 
 * The Original Code is the NRL Parser, released 28 April 2006, 
 * Copyright (c) Christian Nentwich. The Initial Developer of the 
 * Original Code is Christian Nentwich. Portions created by contributors 
 * identified in the NOTICES file are Copyright (c) the contributors. 
 * All Rights Reserved. 
 */
package net.sourceforge.nrl.parser.model.xsd;

import java.io.File;

import junit.framework.TestCase;
import net.sourceforge.nrl.parser.model.IAttribute;
import net.sourceforge.nrl.parser.model.IClassifier;
import net.sourceforge.nrl.parser.model.IModelElement;
import net.sourceforge.nrl.parser.model.IPackage;

import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.xsd.util.XSDResourceFactoryImpl;

/**
 * Test loading of complex types.
 * 
 * @author Christian Nentwich
 */
public class ComplexTypesTest extends TestCase {

	private static boolean isInitialised = false;

	private static IPackage model = null;

	public void setUp() throws Exception {
		if (!isInitialised) {
			Resource.Factory.Registry.INSTANCE.getExtensionToFactoryMap().put("xsd",
					new XSDResourceFactoryImpl());
			isInitialised = true;

			XSDModelLoader loader = new XSDModelLoader();
			model = loader.load(new File("testdata/schema/complextypes.xsd"));

			for (String warning : loader.getWarnings()) {
				System.err.println(warning);
			}
		}
	}

	/*
	 * Just a double-check that the model loaded
	 */
	public void testLoading() throws Exception {
		assertNotNull(model);
	}

	/*
	 * Make sure an "Object" root type was introduced.
	 */
	public void testObjectType() throws Exception {
		IClassifier obj = (IClassifier) model.getElementByName("Object", true);
		assertNotNull(obj);
	}

	/*
	 * Test a flat complex type with a content model based on a sequence.
	 */
	public void testComplexTypeFlat() throws Exception {
		IClassifier classifier = (IClassifier) model.getElementByName("Basic", true);
		assertEquals(IModelElement.ElementType.Classifier, classifier.getElementType());
		
		assertNotNull(classifier);
		assertTrue(!classifier.isEnumeration());
		assertEquals("Object", classifier.getParent().getName());

		assertNotNull(classifier.getAttributeByName("id", false));
		assertEquals(0, classifier.getAttributeByName("id", false).getMinOccurs());
		assertEquals(1, classifier.getAttributeByName("id", false).getMaxOccurs());
		assertEquals("string", classifier.getAttributeByName("id", false).getType()
				.getName());

		assertNotNull(classifier.getAttributeByName("simple", false));
		assertEquals(0, classifier.getAttributeByName("simple", false).getMinOccurs());
		assertEquals(1, classifier.getAttributeByName("simple", false).getMaxOccurs());
		assertEquals("string", classifier.getAttributeByName("simple", false).getType()
				.getName());

		assertNotNull(classifier.getAttributeByName("mySimple", false));
		assertEquals(1, classifier.getAttributeByName("mySimple", false).getMinOccurs());
		assertEquals(1, classifier.getAttributeByName("mySimple", false).getMaxOccurs());
		assertEquals("MySimpleType", classifier.getAttributeByName("mySimple", false)
				.getType().getName());

		assertNotNull(classifier.getAttributeByName("anotherSimple", false));
		assertEquals(1, classifier.getAttributeByName("anotherSimple", false)
				.getMinOccurs());
		assertEquals(IAttribute.UNBOUNDED, classifier.getAttributeByName("anotherSimple",
				false).getMaxOccurs());
		assertEquals("string", classifier.getAttributeByName("anotherSimple", false)
				.getType().getName());
	}

	public void testComplexTypeNesting() throws Exception {
		IClassifier classifier = (IClassifier) model.getElementByName("NestedA", true);

		assertNotNull(classifier);
		assertTrue(!classifier.isEnumeration());

		assertNotNull(classifier.getAttributeByName("childA", false));
		assertEquals(1, classifier.getAttributeByName("childA", false).getMinOccurs());
		assertEquals(1, classifier.getAttributeByName("childA", false).getMaxOccurs());
		assertEquals("string", classifier.getAttributeByName("childA", false).getType()
				.getName());

		assertNotNull(classifier.getAttributeByName("childB", false));
		assertEquals(0, classifier.getAttributeByName("childB", false).getMinOccurs());
		assertEquals(1, classifier.getAttributeByName("childB", false).getMaxOccurs());
		assertEquals("string", classifier.getAttributeByName("childB", false).getType()
				.getName());

		assertNotNull(classifier.getAttributeByName("childC", false));
		assertEquals(0, classifier.getAttributeByName("childC", false).getMinOccurs());
		assertEquals(1, classifier.getAttributeByName("childC", false).getMaxOccurs());
		assertEquals("string", classifier.getAttributeByName("childC", false).getType()
				.getName());

		assertNotNull(classifier.getAttributeByName("childD", false));
		assertEquals(1, classifier.getAttributeByName("childD", false).getMinOccurs());
		assertEquals(1, classifier.getAttributeByName("childD", false).getMaxOccurs());
		assertEquals("string", classifier.getAttributeByName("childD", false).getType()
				.getName());

		assertNotNull(classifier.getAttributeByName("childE", false));
		assertEquals(1, classifier.getAttributeByName("childE", false).getMinOccurs());
		assertEquals(1, classifier.getAttributeByName("childE", false).getMaxOccurs());
		assertEquals("double", classifier.getAttributeByName("childE", false).getType()
				.getName());

		assertNotNull(classifier.getAttributeByName("childF", false));
		assertEquals(1, classifier.getAttributeByName("childF", false).getMinOccurs());
		assertEquals(1, classifier.getAttributeByName("childF", false).getMaxOccurs());
		assertEquals("int", classifier.getAttributeByName("childF", false).getType()
				.getName());

		assertNotNull(classifier.getAttributeByName("childG", false));
		assertEquals(1, classifier.getAttributeByName("childG", false).getMinOccurs());
		assertEquals(1, classifier.getAttributeByName("childG", false).getMaxOccurs());
		assertEquals("long", classifier.getAttributeByName("childG", false).getType()
				.getName());
	}

	public void testComplexAnonymousSubtype() throws Exception {
		IClassifier classifier = (IClassifier) model.getElementByName(
				"WithAnonymousComplex", true);
		assertEquals(IModelElement.ElementType.Classifier, classifier.getElementType());
		
		assertNotNull(classifier);
		assertTrue(!classifier.isEnumeration());

		assertNotNull(classifier.getAttributeByName("simple", false));
		assertEquals(1, classifier.getAttributeByName("simple", false).getMinOccurs());
		assertEquals(1, classifier.getAttributeByName("simple", false).getMaxOccurs());
		assertEquals("string", classifier.getAttributeByName("simple", false).getType()
				.getName());

		assertNotNull(classifier.getAttributeByName("anonymousComplex", false));
		assertEquals(1, classifier.getAttributeByName("anonymousComplex", false)
				.getMinOccurs());
		assertEquals(1, classifier.getAttributeByName("anonymousComplex", false)
				.getMaxOccurs());

		assertNotNull(model.getElementByName("AnonymousComplex", true));
		IClassifier type = (IClassifier) classifier.getAttributeByName(
				"anonymousComplex", false).getType();
		assertEquals("AnonymousComplex", type.getName());
		assertEquals("AnonymousComplex", type.getOriginalName());
		assertEquals("Basic", type.getParent().getName());

		assertNotNull(type.getAttributeByName("simple", true));
		assertNotNull(type.getAttributeByName("mySimple", true));
		assertNotNull(type.getAttributeByName("anotherSimple", true));
		assertNotNull(type.getAttributeByName("test", true));

		assertNotNull(type.getUserData(IXSDUserData.CONTAINING_TYPE));
		assertTrue(type.getUserData(IXSDUserData.CONTAINING_TYPE) == classifier);

		assertNotNull(model.getElementByName("AnonymousComplex2", true));
		type = (IClassifier) classifier.getAttributeByName("anonymousComplex2", false)
				.getType();
		assertEquals("AnonymousComplex2", type.getName());
		assertEquals("AnonymousComplex2", type.getOriginalName());
		assertEquals("Object", type.getParent().getName());

		assertNotNull(type.getAttributeByName("childA", true));
		assertNotNull(type.getAttributeByName("childB", true));

		assertNotNull(type.getUserData(IXSDUserData.CONTAINING_TYPE));
		assertTrue(type.getUserData(IXSDUserData.CONTAINING_TYPE) == classifier);

		classifier = (IClassifier) model.getElementByName("WithAnonymousComplex2", true);
		assertNotNull(classifier.getAttributeByName("anonymousComplex", false));
		type = (IClassifier) classifier.getAttributeByName("anonymousComplex", false)
				.getType();
		assertEquals("AnonymousComplex1", type.getName());
		assertEquals("AnonymousComplex", type.getOriginalName());
		assertEquals("NestedA", type.getParent().getName());

		// Deeply nested
		classifier = (IClassifier) model.getElementByName("AnonymousNested", true);
		assertNotNull(classifier);
		assertNotNull(classifier.getUserData(IXSDUserData.CONTAINING_TYPE));
		assertEquals("AnonymousComplex", ((IClassifier) classifier
				.getUserData(IXSDUserData.CONTAINING_TYPE)).getName());
		assertEquals("WithAnonymousComplex", ((IClassifier) ((IClassifier) classifier
				.getUserData(IXSDUserData.CONTAINING_TYPE))
				.getUserData(IXSDUserData.CONTAINING_TYPE)).getName());

		// anonymous complex type derived from simple
		classifier = (IClassifier) model.getElementByName("WithDerivedAnonymousComplex",
				true);
		assertNotNull(classifier);
		assertNotNull(classifier.getAttributeByName("anonymousComplexElement1", false));
		assertEquals("AnonymousComplexElement1", classifier.getAttributeByName(
				"anonymousComplexElement1", false).getType().getName());

		classifier = (IClassifier) model.getElementByName("AnonymousComplexElement1",
				true);
		assertNotNull(classifier);
		assertNotNull(classifier.getParent());
		assertEquals("normalizedString", classifier.getParent().getName());
		assertNotNull(classifier.getAttributeByName("simpleContentAttribute", false));
		
		classifier = (IClassifier) model.getElementByName("AnonymousComplexElement2",
				true);
		assertNotNull(classifier);
		assertNotNull(classifier.getParent());
		assertEquals("AnotherComplexType", classifier.getParent().getName());
		assertNotNull(classifier.getAttributeByName("complexContentAttribute", false));
	}

	public void testSimpleAnonymousSubtype() throws Exception {
		IClassifier classifier = (IClassifier) model.getElementByName(
				"WithAnonymousSimple", true);

		assertNotNull(classifier);
		assertTrue(!classifier.isEnumeration());

		assertNotNull(classifier.getAttributeByName("facetted", false));
		assertEquals("string", classifier.getAttributeByName("facetted", false).getType()
				.getName());

		assertNotNull(classifier.getAttributeByName("enumerated", false));
		assertEquals("string", classifier.getAttributeByName("enumerated", false)
				.getType().getName());

		assertNotNull(classifier.getAttributeByName("internal", false));
		assertEquals("string", classifier.getAttributeByName("internal", false).getType()
				.getName());
	}

	public void testRepeatingParticles_SimpleCases() throws Exception {
		// Simple repeating sequence
		IClassifier classifier = (IClassifier) model.getElementByName(
				"RepeatingParticle1", true);
		assertNotNull(classifier);

		assertEquals(1, classifier.getAttributes(false).size());
		assertNotNull(classifier.getAttributeByName("elementA", false));
		assertEquals(IAttribute.UNBOUNDED, classifier.getAttributeByName("elementA",
				false).getMaxOccurs());
		assertEquals("string", classifier.getAttributeByName("elementA", false).getType()
				.getName());

		// Simple repeating sequence, spurious intermediate particles
		classifier = (IClassifier) model.getElementByName("RepeatingParticle1b", true);
		assertNotNull(classifier);

		assertEquals(1, classifier.getAttributes(false).size());
		assertNotNull(classifier.getAttributeByName("elementA", false));
		assertEquals(IAttribute.UNBOUNDED, classifier.getAttributeByName("elementA",
				false).getMaxOccurs());
		assertEquals("string", classifier.getAttributeByName("elementA", false).getType()
				.getName());

		// Simple repeating sequence, complex type
		classifier = (IClassifier) model.getElementByName("RepeatingParticle2", true);
		assertNotNull(classifier);

		assertEquals(1, classifier.getAttributes(false).size());
		assertNotNull(classifier.getAttributeByName("elementA", false));
		assertEquals(IAttribute.UNBOUNDED, classifier.getAttributeByName("elementA",
				false).getMaxOccurs());
		assertEquals("Basic", classifier.getAttributeByName("elementA", false).getType()
				.getName());

		// Sequence of mixed type elements
		classifier = (IClassifier) model.getElementByName("RepeatingParticle3", true);
		assertNotNull(classifier);

		assertEquals(1, classifier.getAttributes(false).size());
		assertNotNull(classifier.getAttributeByName("elementAAndElementB", false));
		assertEquals(IAttribute.UNBOUNDED, classifier.getAttributeByName(
				"elementAAndElementB", false).getMaxOccurs());
		assertEquals("Object", classifier
				.getAttributeByName("elementAAndElementB", false).getType().getName());

		classifier = (IClassifier) model.getElementByName("RepeatingParticle3b", true);
		assertNotNull(classifier);

		assertEquals(1, classifier.getAttributes(false).size());
		assertNotNull(classifier.getAttributeByName("elementAAndElementB", false));
		assertEquals(IAttribute.UNBOUNDED, classifier.getAttributeByName(
				"elementAAndElementB", false).getMaxOccurs());
		assertEquals(0, classifier.getAttributeByName("elementAAndElementB", false)
				.getMinOccurs());
		assertEquals("Object", classifier
				.getAttributeByName("elementAAndElementB", false).getType().getName());

		// Sequence of same complex type
		classifier = (IClassifier) model.getElementByName("RepeatingParticle4", true);
		assertNotNull(classifier);

		assertEquals(1, classifier.getAttributes(false).size());
		assertNotNull(classifier.getAttributeByName("elementAAndElementB", false));
		assertEquals(IAttribute.UNBOUNDED, classifier.getAttributeByName(
				"elementAAndElementB", false).getMaxOccurs());
		assertEquals("ElementAAndElementB", classifier.getAttributeByName(
				"elementAAndElementB", false).getType().getName());

		// Check the artificially introduced classifier
		classifier = (IClassifier) model.getElementByName("ElementAAndElementB", true);
		assertEquals(2, classifier.getAttributes(false).size());
		assertNotNull(classifier.getAttributeByName("elementA", false));
		assertEquals("Basic", classifier.getAttributeByName("elementA", false).getType()
				.getName());
		assertEquals(1, classifier.getAttributeByName("elementA", false).getMaxOccurs());
		assertNotNull(classifier.getAttributeByName("elementB", false));
		assertEquals("Basic", classifier.getAttributeByName("elementB", false).getType()
				.getName());
		assertEquals(1, classifier.getAttributeByName("elementB", false).getMaxOccurs());

		// Choice of the same complex type
		classifier = (IClassifier) model.getElementByName("RepeatingParticle5", true);
		assertNotNull(classifier);

		assertEquals(1, classifier.getAttributes(false).size());
		assertNotNull(classifier.getAttributeByName("elementAOrElementB", false));
		assertEquals(IAttribute.UNBOUNDED, classifier.getAttributeByName(
				"elementAOrElementB", false).getMaxOccurs());
		assertEquals("Object", classifier.getAttributeByName("elementAOrElementB", false)
				.getType().getName());

		// More than two elements within a repeating sequence
		classifier = (IClassifier) model.getElementByName("RepeatingParticle6", true);
		assertNotNull(classifier);
		assertEquals(1, classifier.getAttributes(false).size());
		assertNotNull(classifier.getAttributeByName("elementAAndElementBOrElementC",
				false));
		assertEquals(IAttribute.UNBOUNDED, classifier.getAttributeByName(
				"elementAAndElementBOrElementC", false).getMaxOccurs());
		assertEquals("ElementAAndElementBOrElementC", classifier.getAttributeByName(
				"elementAAndElementBOrElementC", false).getType().getName());

		// Check the artificially introduced classifier
		classifier = (IClassifier) model.getElementByName(
				"ElementAAndElementBOrElementC", true);
		assertEquals(4, classifier.getAttributes(false).size());
		assertNotNull(classifier.getAttributeByName("elementA", false));
		assertNotNull(classifier.getAttributeByName("elementB", false));
		assertNotNull(classifier.getAttributeByName("elementC", false));
		assertNotNull(classifier.getAttributeByName("elementD", false));

		// More than two elements, with spurious particles
		classifier = (IClassifier) model.getElementByName("RepeatingParticle7", true);
		assertNotNull(classifier);
		assertEquals(1, classifier.getAttributes(false).size());
		assertNotNull(classifier.getAttributeByName("elementAOrElementBAndElementC",
				false));
		assertEquals(IAttribute.UNBOUNDED, classifier.getAttributeByName(
				"elementAOrElementBAndElementC", false).getMaxOccurs());
		assertEquals("ElementAOrElementBAndElementC", classifier.getAttributeByName(
				"elementAOrElementBAndElementC", false).getType().getName());

		// Check the artificially introduced classifier
		classifier = (IClassifier) model.getElementByName(
				"ElementAOrElementBAndElementC", true);
		assertEquals(3, classifier.getAttributes(false).size());
		assertNotNull(classifier.getAttributeByName("elementA", false));
		assertNotNull(classifier.getAttributeByName("elementB", false));
		assertNotNull(classifier.getAttributeByName("elementC", false));

		// More element content, with spurious sequences
		classifier = (IClassifier) model.getElementByName("RepeatingParticle8", true);
		assertNotNull(classifier);
		assertEquals(1, classifier.getAttributes(false).size());
		assertNotNull(classifier.getAttributeByName("aAndBAndC", false));
		assertEquals(IAttribute.UNBOUNDED, classifier.getAttributeByName("aAndBAndC",
				false).getMaxOccurs());
		assertEquals("AAndBAndC", classifier.getAttributeByName("aAndBAndC", false)
				.getType().getName());

		// Check the artificially introduced classifier
		classifier = (IClassifier) model.getElementByName("AAndBAndC", true);
		assertEquals(6, classifier.getAttributes(false).size());

		// With reference to global element
		classifier = (IClassifier) model.getElementByName("RepeatingParticle9", true);
		assertNotNull(classifier);
		assertEquals(1, classifier.getAttributes(false).size());
		assertNotNull(classifier.getAttributeByName("elementOrB", false));
		assertEquals(IAttribute.UNBOUNDED, classifier.getAttributeByName("elementOrB",
				false).getMaxOccurs());
		assertEquals("Object", classifier.getAttributeByName("elementOrB", false)
				.getType().getName());
	}
}
